@@ -1,5 +1,5 @@
 """
-Implementation of the XDG Menu Specification Version 0.7
+Implementation of the XDG Menu Specification Version 0.8
 http://www.freedesktop.org/standards/menu-spec
 
 Not Supported (and not planed):
@@ -7,7 +7,7 @@ Not Supported (and not planed):
 """
 
 from __future__ import generators
-import re, os, xml.dom.minidom
+import os, xml.dom.minidom
 
 from xdg.BaseDirectory import *
 from xdg.DesktopEntry import *
@@ -25,6 +25,7 @@ class Menu:
 		self.DefaultLayout = ""
 		self.Deleted = "notset"
 		self.DeskEntries = []
+		self.CacheDeskEntries = []
 		self.Directory = []
 		self.DirectoryDirs = []
 		self.Layout = ""
@@ -78,8 +79,8 @@ class Menu:
 	def setWM(self, wm):
 		for entry in self.Entries:
 			if isinstance(entry, MenuEntry):
-				if ( entry.DesktopEntry.hasKey("OnlyShowIn") and not wm in entry.DesktopEntry.getOnlyShowIn()) \
-				or ( entry.DesktopEntry.hasKey("NotShowIn") and wm in entry.DesktopEntry.getNotShowIn()):
+				if wm not in entry.DesktopEntry.getOnlyShowIn() \
+				or wm in entry.DesktopEntry.getNotShowIn():
 					entry.Show = False
 				else:
 					entry.Show = True
@@ -90,7 +91,7 @@ class Menu:
 		for entry in self.Entries:
 			if isinstance(entry, MenuEntry):
 				entry.DesktopEntry.setLocale(lc_messages)
-				entry.Name = entry.DesktopEntry.getName()
+				entry.cache()
 			elif isinstance(entry, Menu):
 				level += 1
 				entry.Directory.setLocale(lc_messages)
@@ -193,7 +194,9 @@ class Menu:
 		return self.Deleted
 
 	def addDeskEntry(self, entry):
-		if not entry in self.getDeskEntries():
+		#if not entry in self.DeskEntries:
+		if not entry.Name in self.CacheDeskEntries:
+			self.CacheDeskEntries.append(entry.Name)
 			self.DeskEntries.append(entry)
 	def getDeskEntries(self):
 		return self.DeskEntries
@@ -358,10 +361,25 @@ class Rule:
 
 		# Begin parsing
 		self.parseNode(node)
+		self.compile()
 
 	def __str__(self):
 		return self.Rule
 
+	def compile(self):
+		exec("""
+def do(entries, type):
+    for entry in entries:
+		if entry.Allocated:
+		    continue
+		if %s:
+		    if type == "Include":
+				entry.Add = True
+		    else:
+				entry.Add = False
+    return entries
+""" % self.Rule) in self.__dict__
+
 	def parseNode(self, node):
 		for child in node.childNodes:
 			if child.nodeType == xml.dom.Node.ELEMENT_NODE:
@@ -392,7 +410,7 @@ class Rule:
 
 	def parseCategory(self, value):
 		self.parseNew()
-		self.Rule += "'" + value + "' in categories"
+		self.Rule += "'" + value + "' in entry.Categories"
 
 	def parseAll(self):
 		self.parseNew()
@@ -431,15 +449,22 @@ class Rule:
 
 class MenuEntry:
 	"Wrapper for 'Menu Style' Desktop Entries"
-	def __init__(self, Entry = "", Id = "", Allocated = False):
+	def __init__(self, Entry, Id = "", Allocated = False):
 		self.DesktopEntry = Entry
 		self.DesktopFileID = Id
 		self.Allocated = Allocated
+		self.Add = False
 		# to implement the OnlyShowIn/NotShowIn keys
 		self.Show = True
 		# Caching
-		if self.DesktopEntry:
-			self.Name = self.DesktopEntry.getName()
+		self.Name = ""
+		self.Categories = ""
+		self.cache()
+	
+	def cache(self):
+		self.Name = self.DesktopEntry.getName()
+		if not self.Categories:
+			self.Categories = self.DesktopEntry.getCategories()
 
 	def __cmp__(self, other):
 		if isinstance(other, MenuEntry):
@@ -506,7 +531,7 @@ def parse(file = ""):
 		file = os.path.abspath(file)
 
 	# check if it is a .menu file
-	if not re.search('\.menu$', file):
+	if not os.path.splitext(file)[1] == ".menu":
 		raise ParsingError('Not a .menu file', file)
 
 	# create xml parser
@@ -537,7 +562,7 @@ def parse(file = ""):
 def __preparse(doc, file):
 	# extrace path and basename from file
 	path = os.path.dirname(file)
-	(basename, nil) = re.split(".[^.]*$", os.path.basename(file))
+	basename = os.path.splitext(os.path.basename(path))[0]
 
 	# replace default dir stuff
 	for name in [ "DefaultAppDirs", "DefaultDirectoryDirs", "DefaultMergeDirs" ]:
@@ -558,7 +583,7 @@ def __preparse(doc, file):
 				elif name == "DefaultMergeDirs":
 					value = os.path.join(dir, "menus", basename + "-merged")
 
-				entry = doc.createElement(re.sub("Default", "", name).replace("Dirs", "Dir"))
+				entry = doc.createElement(name.replace("Default", "").replace("Dirs", "Dir"))
 				entry.appendChild(doc.createTextNode(value))
 				parent.insertBefore(entry, node)
 			parent.removeChild(node)
@@ -706,7 +731,7 @@ def __parseMergeFile(child):
 def __parseMergeDir(child):
 	files = os.listdir(child.childNodes[0].nodeValue)
 	for file in files:
-		if re.search('\.menu', file):
+		if os.path.splitext(file)[1] == ".menu":
 			__mergeFile(os.path.join(child.childNodes[0].nodeValue, file), child)
 
 def __mergeFile(file, child):
@@ -743,16 +768,12 @@ def __parseMergeEnd():
 def __genmenuNotOnlyAllocated(menu, cache):
 	if not menu.getOnlyUnallocated():
 		cache.addEntries(menu.getAppDirs())
-		for entry in cache.getEntries(menu.getAppDirs()):
-			categories = entry.DesktopEntry.getCategories()
-			inc = False
-			for rule in menu.getRules():
-				if eval(rule.Rule):
-					if rule.Type == "Include":
-						inc = True
-					else:
-						inc = False
-			if inc == True:
+		entries = []
+		for rule in menu.getRules():
+			entries = rule.do(cache.getEntries(menu.getAppDirs()), rule.Type)
+		for entry in entries:
+		    if entry.Add == True:
+				entry.Add = False
 				entry.Allocated = True
 				menu.addDeskEntry(entry)
 	for submenu in menu.getSubmenus():
@@ -761,18 +782,13 @@ def __genmenuNotOnlyAllocated(menu, cache):
 def __genmenuOnlyAllocated(menu, cache):
 	if menu.getOnlyUnallocated():
 		cache.addEntries(menu.getAppDirs())
-		for entry in cache.getEntries(menu.getAppDirs()):
-			if entry.Allocated == True:
-				continue
-			categories = entry.DesktopEntry.getCategories()
-			inc = False
-			for rule in menu.Rules:
-				if eval(rule.Rule):
-					if rule.Type == "Include":
-						inc = True
-					else:
-						inc = False
-			if inc == True:
+		entries = []
+		for rule in menu.getRules():
+			entries = rule.do(cache.getEntries(menu.getAppDirs()), rule.Type)
+		for entry in entries:
+		    if entry.Add == True:
+			#	entry.Add = False
+			#	entry.Allocated = True
 				menu.addDeskEntry(entry)
 
 	for submenu in menu.getSubmenus():
@@ -861,6 +877,7 @@ class DesktopEntryCache:
 	"Class to cache Desktop Entries"
 	def __init__(self):
 		self.cacheEntries = {}
+		self.cache = {}
 
 	def addEntries(self, dirs):
 		for dir in dirs:
@@ -871,16 +888,16 @@ class DesktopEntryCache:
 	def __addFiles(self, dir, subdir):
 		files = os.listdir(os.path.join(dir,subdir))
 		for file in files:
-			if re.search('\.desktop', file):
+			if os.path.splitext(file)[1] == ".desktop":
 				try:
 					deskentry = DesktopEntry()
 					deskentry.parse(os.path.join(dir, subdir, file))
 					if not deskentry.getHidden() \
 					and not deskentry.getNoDisplay():
-						entry = MenuEntry(deskentry, re.sub("/", "-", os.path.join(subdir,file)))
+						entry = MenuEntry(deskentry, os.path.join(subdir,file).replace("/", "-"))
 						self.cacheEntries[dir].append(entry)
-				except:
-					pass
+				except ParsingError:
+					continue
 			elif os.path.isdir(os.path.join(dir,subdir,file)):
 				self.__addFiles(dir, os.path.join(subdir,file))
 
@@ -888,10 +905,17 @@ class DesktopEntryCache:
 		list = []
 		# don't compare MenuEntries, this is veryyyy slow, cache their DesktopFileIDs
 		ids = []
+		# cache the results again
+		key = "".join(dirs)
+		try:
+			return self.cache[key]
+		except KeyError:
+			pass
 		for dir in dirs:
 			if self.cacheEntries.has_key(dir):
 				for entry in self.cacheEntries[dir]:
 					if not entry.DesktopFileID in ids:
 						ids.append(entry.DesktopFileID)
 						list.append(entry)
+		self.cache[key] = list
 		return list
